import os
import logging
import sys
import time
from os.path import isdir, join

from aspen import restarter
from aspen.website import Website
from aspen.configuration import Configuration
from diesel import Application, Loop, Service
from diesel.protocols import http


log = logging.getLogger('aspen.cli')


def main(argv=None):
    try:
        configuration = Configuration(argv)
        configuration.app = app = Application()
        website = Website(configuration)
        configuration.website = website # to support re-handling, especially
        website = configuration.hooks.run('startup', website)

        # change current working directory
        os.chdir(configuration.root)

        port = configuration.address[1]
        app.add_service(Service(http.HttpServer(website), port))

        log.warn("Greetings, program! Welcome to port %d." % port)
        app.run()

    except KeyboardInterrupt, SystemExit:
        configuration.hooks.run('shutdown', website)

def thrash():
    """This is a very simple tool to restart a process when it dies.

    It's designed to restart aspen in development when it dies because files
    have changed and you set changes_kill in the [aspen] section of aspen.conf.

    """
    try:
        while 1:
            os.system(' '.join(sys.argv[1:]))
            time.sleep(1)
    except KeyboardInterrupt:
        pass



"""Define configuration objects.
"""
import logging
import logging.config
import mimetypes
import os
import socket
import sys
from logging import StreamHandler
from logging.handlers import TimedRotatingFileHandler
from os.path import dirname, exists, expanduser, isdir, join, realpath

import aspen
from aspen.configuration.aspenconf import AspenConf 
from aspen.configuration.exceptions import ConfigurationError
from aspen.configuration.hooks import HooksConf
from aspen.configuration.optparser import ( optparser
                                          , validate_address
                                          , validate_log_level
                                           )
from aspen._tornado.template import Loader


log = logging.getLogger('aspen') # configured below; not used until then


LOG_FORMAT = "%(message)s"
LOG_LEVEL = logging.WARNING


class Configuration(object):
    """Aggregate configuration from several sources.
    """

    root = ''       # string, filesystem root
    args = None     # argument list as returned by OptionParser.parse_args
    conf = None     # a AspenConf instance
    optparser = None # an optparse.OptionParser instance
    opts = None     # an optparse.Values instance per OptionParser.parse_args

    address = None  # the AF_INET, AF_INET6, or AF_UNIX address to bind to
    sockfam = None  # one of socket.AF_{INET,INET6,UNIX}


    def __init__(self, argv):
        """Takes an argv list, and gives it straight to optparser.parse_args.
        """

        # Initialize parsers.
        # ===================

        opts, args = optparser.parse_args(argv)


        # Root
        # ====
        # This can only be passed on the command line.

        if args:
            root = args[0]
        else:
            try:
                # Under supervisord, the following raises 
                #   OSError: [Errno 2] No such file or directory
                # So be sure to pass a directory in on the command line, or cwd
                # using supervisord's own facility for that.
                root = os.getcwd()
            except OSError:
                raise ConfigurationError("Could not get a current working "
                                         "directory. You can specify the site "
                                         "root on the command line.")
        root = realpath(root)
        if not isdir(root):
            msg = "%s does not point to a directory" % root
            raise ConfigurationError(msg)


        # sys.path
        # ========

        dotaspen = join(root, '.aspen')
        if isdir(dotaspen):
            sys.path.insert(0, dotaspen)


        # aspen.conf
        # ==========

        conf = AspenConf( '/etc/aspen/aspen.conf'
                        , '/usr/local/etc/aspen/aspen.conf'
                        , expanduser('~/.aspen/aspen.conf') 
                        , join(dotaspen, 'aspen.conf')
                         ) # later overrides earlier
        
        self.root = root
        self.args = args
        self.conf = conf
        self.optparser = optparser
        self.opts = opts


        # Loader
        # ======

        self.loader = Loader(dotaspen)


        # mimetypes
        # =========
        
        mimetypes.knownfiles = [ join(dirname(__file__), 'mime.types')
                               , '/etc/mime.types'
                               , '/usr/local/etc/mime.types'
                               , join(dotaspen, 'mime.types')
                                ]
        mimetypes.init()

        self.default_mimetype = conf.aspen.get( 'default_mimetype'
                                              , 'text/plain'
                                               )

        # index.html
        # ==========

        default_filenames = conf.aspen.get('default_filenames', 'index.html')
        default_filenames = default_filenames.split()
        default_filenames = [x.strip(',') for x in default_filenames]
        default_filenames = [x for x in default_filenames if x]
        default_filenames = [x.split(',') for x in default_filenames]
        self.default_filenames = []
        for nested in default_filenames:
            self.default_filenames.extend(nested)

       
        # hooks
        # =====

        self.hooks = HooksConf( join(dirname(__file__), 'hooks.conf')
                              , '/etc/aspen/hooks.conf'
                              , '/usr/local/etc/aspen/hooks.conf'
                              , expanduser('~/.aspen/hooks.conf')
                              , join(dotaspen, 'hooks.conf')
                               ) # later comes after earlier, per section


        # address/sockfam
        # ===============
        # These can be set either on the command line or in the conf file.

        if getattr(opts, 'have_address', False):        # first check CLI
            address = opts.address
            sockfam = opts.sockfam
        elif 'address' in conf.main:                    # then check conf
            address, sockfam = validate_address(conf.main['address'])
        else:                                           # default from optparse
            address = opts.address
            sockfam = socket.AF_INET

        self.address = address
        self.sockfam = sockfam


        # Logging
        # =======
        # Logging can be configured from four places, in this order of 
        # precedence:
        #
        #   1) some other module (theoretically, I'm not sure an aspen user 
        #      could make this happen easily)
        #   2) the command line
        #   3) a logging.conf file
        #   4) a [logging] section in the aspen.conf file
        #
        # These are not layered; only one is used.

        #FMT = "%(process)-6s%(levelname)-9s%(name)-14s%(message)s"

        logging_configured = bool(len(logging.getLogger().handlers))
                             # this test is taken from logging.basicConfig.

        if logging_configured:              # some other module
            log.warn("logging is already configured")

        if not logging_configured:          # command line
            kw = dict()
            kw['filename'] = opts.log_file
            kw['filter'] = opts.log_filter
            kw['format'] = opts.log_format
            kw['level'] = opts.log_level
            if kw.values() != [None, None, None, None]: # at least one knob set
                if kw['format'] is None:
                    kw['format'] = LOG_FORMAT
                if kw['level'] is None:
                    kw['level'] = LOG_LEVEL
                self.configure_logging(**kw)
                log.info("logging configured from the command line")
                logging_configured = True

        if not logging_configured:          # logging.conf
            # TODO /etc/aspen/logging.conf
            # TODO /usr/local/etc/aspen/logging.conf
            logging_conf = join(root, '.aspen', 'logging.conf')
            if exists(logging_conf):
                logging.config.fileConfig(logging_conf) 
                log.info("logging configured from logging.conf")
                logging_configured = True

        if not logging_configured:          # aspen.conf [logging]
            kw = dict()
            kw['filename'] = conf.logging.get('file')
            kw['filter'] = conf.logging.get('filter')
            kw['format'] = conf.logging.get('format', LOG_FORMAT)

            log_level = conf.logging.get('level')
            if log_level is not None:
                log_level = validate_log_level(log_level) 
            else:
                log_level = LOG_LEVEL
            kw['level'] = log_level

            self.configure_logging(**kw)
            log.info("logging configured from aspen.conf")
            logging_configured = True


    def configure_logging(self, filename, filter, format, level):
        """Used for configuring logging from the command line or aspen.conf.
        """
   
        # Handler
        # =======
        # sys.stdout or rotated file
   
        if filename is None:
            handler = StreamHandler(sys.stdout)
        else:
            # @@: Handle absolute paths on Windows
            #  http://sluggo.scrapping.cc/python/unipath/Unipath-current/unipath/abstractpath.py
            #  http://docs.python.org/library/os.path.html#os.path.splitunc
            if not filename.startswith('/'):
                filename = join(self.paths.root, filename)
                filename = realpath(filename)
            logdir = dirname(filename)
            if not isdir(logdir):
                os.makedirs(logdir, 0755)
            handler = TimedRotatingFileHandler( filename=filename
                                              , when='midnight'
                                              , backupCount=7
                                               )
        # Filter
        # ======
        
        if filter is not None:
            filter = logging.Filter(filter)
            handler.addFilter(filter)
    
    
        # Format
        # ======
   
        formatter = logging.Formatter(fmt=format)
        handler.setFormatter(formatter)


        # Level
        # =====

        handler.setLevel(level)

    
        # Installation
        # ============
   
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        root_logger.setLevel(level) # bah



"""Simplates

Problems with tornado.template:

    - no option to fail silently
    - Loader cache doesn't account for modtime
    - Is this a bug?

        {{ foo }}
        {% for foo in [1,2,3] %}
        {% end %}

    - no loop counters, eh? must do it manually with {% set %}

"""
import logging
import mimetypes
import os
import stat
import sys
import traceback
from os.path import join

import diesel
from aspen.http import Response
from _tornado.template import Loader, Template


PAGE_BREAK = chr(12)
log = logging.getLogger('aspen.simplate')


class LoadError(StandardError):
    """Represent a problem parsing a simplate.
    """


# Cache helpers
# =============

__cache = dict()        # cache, keyed to filesystem path

class Entry:
    """An entry in the global simplate cache.
    """

    fspath = ''         # The filesystem path [string]
    modtime = None      # The timestamp of the last change [int]
    quadruple = None    # A post-processed version of the data [4-tuple]
    exc = None          # Any exception in reading or compilation [Exception]

    def __init__(self):
        self.fspath = ''
        self.modtime = 0
        self.quadruple = ()


# Core loader
# ===========

def load_uncached(request):
    """Given a Request object, return three objects (uncached).

    A simplate is a template with two optional Python components at the head of
    the file, delimited by '^L'. The first Python page is exec'd when the
    simplate is first called, and the namespace it populates is saved for all
    subsequent runs. The second Python page is exec'd within the template
    namespace each time the template is rendered.

    If the mimetype does not start with 'text/', then it is only a simplate if
    it has at least one form feed in it. Binary files generally can't be
    decoded using UTF-8. If Python's mimetypes module doesn't know about a
    certain extension, then we default to a configurable value (default is
    text/plain).

    """

    simplate = open(request.fs).read()
    
    # We work with simplate exclusively as a bytestring. Any unicode objects
    # passed in by the user as {{ expressions }} will be encoded with UTF-8 by
    # Tornado.
    
    mimetype = mimetypes.guess_type(request.fs, strict=False)[0]
    if mimetype is None:
        mimetype = request.configuration.default_mimetype
   
    if mimetype.startswith('text/'):
        pass # TODO Consider exiting early if there are no ^L, , {% or {{ 
             # characters. Is this actually faster?
             # https://github.com/whit537/aspen/issues/1
    else:
        s = lambda s: simplate.startswith(s)
        if not (s('#!') or s('"""') or s('import') or s('from')):
            # I tried checking for 1 or 2 form feeds, but found a binary file
            # in the wild that indeed had exactly two. And now we may have
            # caret-L instead, anyway. Let's try sniffing the first few bytes.
            return (mimetype, None, None, simplate) # static file; exit early

    simplate = simplate.replace("^L", PAGE_BREAK)

    npage_breaks = simplate.count(PAGE_BREAK)
    if npage_breaks == 0:
        script = imports = ""
        template = simplate
    elif npage_breaks == 1:
        imports = ""
        script, template = simplate.split(PAGE_BREAK)
    elif npage_breaks == 2:
        imports, script, template = simplate.split(PAGE_BREAK)
    else:
        raise SyntaxError( "Simplate %s may have at most two " % request.fs
                         + "page breaks; it has %d." % npage_breaks
                          )


    # Standardize newlines.
    # =====================
    # compile requires \n, and doing it now makes the next line easier.

    imports = imports.replace('\r\n', '\n')
    script = script.replace('\r\n', '\n')


    # Pad the beginning of the second page.
    # =====================================
    # This is so we get accurate tracebacks. We used to do this for the
    # template page too, but Tornado templates have some weird error handling
    # that we haven't exposed yet.

    script = ''.join(['\n' for n in range(imports.count('\n'))]) + script


    # Prep our cachable objects and return.
    # =====================================

    namespace = dict()
    namespace['__file__'] = request.fs
    namespace['website'] = request.website
    script = compile(script, request.fs, 'exec')
    if template.strip():
        template = Template( template
                           , name = request.fs
                           , loader = request.website.loader
                           , compress_whitespace = False
                            )
    else:
        template = None

    exec compile(imports, request.fs, 'exec') in namespace

    return (mimetype, namespace, script, template)


# Cache wrapper
# =============

def load(request):
    """Given a Request object, return four objects (with caching).
    """

    # Get a cache Entry object.
    # =========================

    if request.fs not in __cache:
        entry = Entry()
        #entry.fspath = request.fs # TODO: What was this for?
        __cache[request.fs] = entry

    entry = __cache[request.fs]


    # Process the simplate.
    # =====================

    modtime = os.stat(request.fs)[stat.ST_MTIME]
    if entry.modtime == modtime:                            # cache hit
        if entry.exc is not None:
            raise entry.exc
    else:                                                   # cache miss
        try:
            entry.quadruple = load_uncached(request)
        except Exception, exception:
            # NB: Old-style string exceptions will still raise.
            entry.exc = ( LoadError(traceback.format_exc())
                        , sys.exc_info()[2]
                         )
        else: # reset any previous exception
            entry.exc = None 

        entry.modtime = modtime
        if entry.exc is not None:
            raise entry.exc[0] # TODO Why [0] here, and not above?


    # Return
    # ======
    # Avoid mutating the cached namespace dictionary.

    mimetype, namespace, script, template = entry.quadruple
    if namespace is not None:
        namespace = namespace.copy()
    return (mimetype, namespace, script, template)


# Main callable.
# ==============

def handle(request, response=None):
    """Given a Request, return or raise a Response.
    """
    if response is None:
        response = Response()

    mimetype, namespace, script, template = load(request)

    if namespace is None:
        response.body = template
    else:
       
        # Populate namespace.
        # ===================
    
        namespace.update(request.namespace)
        namespace['request'] = request
        namespace['response'] = response
   

        # Exec the script.
        # ================
    
        if script:
            exec script in namespace
            response = namespace['response']


        # Process the template.
        # =====================
        # If template is None that means that that page was empty.
    
        if template is not None:
            response.body = template.generate(**namespace)


    # Set the mimetype.
    # =================
    # Note that we guess based on the filesystem path, not the URL path.
    
    if response.headers.one('Content-Type') is None:
        if mimetype.startswith('text/'):
            mimetype += "; charset=UTF-8" 
        response.headers.set('Content-Type', mimetype)


    # Send it on back up the line.
    # ============================

    return response



from aspen.http.mapping import Mapping
from diesel.protocols.http import HttpHeaders as DieselHeaders


class Headers(Mapping):
    """Represent the headers in an HTTP Request message.
    """

    def __init__(self, headers):
        """Takes headers as a string.
        """
        diesel_headers = DieselHeaders()
        diesel_headers.parse(headers)
        self._diesel_headers = diesel_headers
        Mapping.__init__(self)
        self._dict.update(diesel_headers._headers)

    def to_http(self):
        """Return the headers as a string, formatted for an HTTP message.
        """
        out = []
        for header, values in self._dict.iteritems():
            for value in values:
                out.append('%s: %s' % (header.title(), value))
        return '\r\n'.join(out)


